@@ -1,9 +1,10 @@
-#!/usr/bin/env python3
-
 import hashlib
+import logging
 import os
+import shutil
+import subprocess
 import sys
-import types
+import collections
 
 
 DAZEL_RC_FILE = ".dazelrc"
@@ -13,26 +14,36 @@ BAZEL_WORKSPACE_FILE = "WORKSPACE"
 DEFAULT_INSTANCE_NAME = "dazel"
 DEFAULT_IMAGE_NAME = "dazel"
 DEFAULT_RUN_COMMAND = "/bin/bash"
+DEFAULT_DOCKER_COMMAND = "docker"
 DEFAULT_LOCAL_DOCKERFILE = "Dockerfile.dazel"
-DEFAULT_REMOTE_RPOSITORY = "dazel"
+DEFAULT_REMOTE_REPOSITORY = "dazel"
 DEFAULT_DIRECTORY = os.getcwd()
 DEFAULT_COMMAND = "/usr/bin/bazel"
 DEFAULT_VOLUMES = []
 DEFAULT_PORTS = []
+DEFAULT_ENV_VARS = []
 DEFAULT_NETWORK = "dazel"
 DEFAULT_RUN_DEPS = []
 DEFAULT_DOCKER_COMPOSE_FILE = ""
+DEFAULT_DOCKER_COMPOSE_COMMAND = "docker-compose"
 DEFAULT_DOCKER_COMPOSE_PROJECT_NAME = "dazel"
 DEFAULT_DOCKER_COMPOSE_SERVICES = ""
+DEFAULT_USER = ""
+DEFAULT_DOCKER_BUILD_ARGS = ""
 
-DEFAULT_BAZEL_USER_OUTPUT_ROOT = ("%s/.cache/bazel/_bazel_%s" %
-                                  (os.environ.get("HOME", "~"),
-                                   os.environ.get("USER", "user")))
+DEFAULT_DELEGATED_VOLUME = True
+DEFAULT_BAZEL_USER_OUTPUT_ROOT = os.path.expanduser("~/.cache/bazel/_bazel_%s" %
+                                                    os.environ.get("USER", "user"))
 TEMP_BAZEL_OUTPUT_USER_ROOT = ("/var/bazel/workspace/_bazel_%s" %
                                os.environ.get("USER", "user"))
 DEFAULT_BAZEL_USER_OUTPUT_PATHS = ["external", "action_cache", "execroot"]
 DEFAULT_BAZEL_RC_FILE = ""
 DEFAULT_DOCKER_RUN_PRIVILEGED = False
+DEFAULT_DOCKER_MACHINE = None
+DEFAULT_WORKSPACE_HEX = False
+
+
+logger = logging.getLogger("dazel")
 
 
 class DockerInstance:
@@ -42,36 +53,54 @@ class DockerInstance:
     set it up through configuration variables, and pass on commands to it.
     It streams the output directly and blocks until the command finishes.
     """
-    
-    def __init__(self, instance_name, image_name, run_command, dockerfile,
-                       repository, directory, command, volumes, ports, network,
-                       run_deps, docker_compose_file, docker_compose_project_name,
-                       docker_compose_services, bazel_user_output_root, bazel_rc_file,
-                       docker_run_privileged, dazel_run_file):
+
+    def __init__(self, instance_name, image_name, run_command, docker_command, dockerfile,
+                       repository, directory, command, volumes, ports, env_vars, network,
+                       run_deps, docker_compose_file, docker_compose_command,
+                       docker_compose_project_name, docker_compose_services, bazel_user_output_root,
+                       bazel_rc_file, docker_run_privileged, docker_machine, dazel_run_file,
+                       workspace_hex, delegated_volume, user, docker_build_args):
+        real_directory = os.path.realpath(directory)
+        self.workspace_hex_digest = ""
         self.instance_name = instance_name
         self.image_name = image_name
         self.run_command = run_command
+        self.docker_command = docker_command
         self.dockerfile = dockerfile
         self.repository = repository
         self.directory = directory
         self.command = command
         self.network = network
         self.docker_compose_file = docker_compose_file
+        self.docker_compose_command = docker_compose_command
         self.docker_compose_project_name = docker_compose_project_name
         self.bazel_user_output_root = bazel_user_output_root
         self.bazel_output_base = ""
         self.bazel_rc_file = bazel_rc_file
         self.docker_run_privileged = docker_run_privileged
+        self.docker_machine = docker_machine
         self.dazel_run_file = dazel_run_file
+        self.delegated_volume_flag = ":delegated" if delegated_volume else ""
+        self.user = user
+        self.docker_build_args = docker_build_args
+
+        if workspace_hex:
+            self.workspace_hex_digest = hashlib.md5(real_directory.encode("ascii")).hexdigest()
+            self.instance_name = "%s_%s" % (self.instance_name, self.workspace_hex_digest)
+            self.docker_compose_project_name = "%s%s" % (self.docker_compose_project_name,
+                                                         self.workspace_hex_digest)
+            if os.path.exists(self.dockerfile):
+                self.image_name = "%s_%s" % (self.image_name, self.workspace_hex_digest)
 
         if self.docker_compose_file:
-            self.network = "%s_%s" % (docker_compose_project_name, network)
+            self.network = "%s_%s" % (self.docker_compose_project_name, network)
 
         self._add_volumes(volumes)
         self._add_ports(ports)
+        self._add_env_vars(env_vars)
         self._add_run_deps(run_deps)
         self._add_compose_services(docker_compose_services)
-        
+
     @classmethod
     def from_config(cls):
         config = cls._config_from_file()
@@ -80,16 +109,20 @@ class DockerInstance:
                 instance_name=config.get("DAZEL_INSTANCE_NAME", DEFAULT_INSTANCE_NAME),
                 image_name=config.get("DAZEL_IMAGE_NAME", DEFAULT_IMAGE_NAME),
                 run_command=config.get("DAZEL_RUN_COMMAND", DEFAULT_RUN_COMMAND),
+                docker_command=config.get("DAZEL_DOCKER_COMMAND", DEFAULT_DOCKER_COMMAND),
                 dockerfile=config.get("DAZEL_DOCKERFILE", DEFAULT_LOCAL_DOCKERFILE),
-                repository=config.get("DAZEL_REPOSITORY", DEFAULT_REMOTE_RPOSITORY),
+                repository=config.get("DAZEL_REPOSITORY", DEFAULT_REMOTE_REPOSITORY),
                 directory=config.get("DAZEL_DIRECTORY", DEFAULT_DIRECTORY),
                 command=config.get("DAZEL_COMMAND", DEFAULT_COMMAND),
                 volumes=config.get("DAZEL_VOLUMES", DEFAULT_VOLUMES),
                 ports=config.get("DAZEL_PORTS", DEFAULT_PORTS),
+                env_vars=config.get("DAZEL_ENV_VARS", DEFAULT_ENV_VARS),
                 network=config.get("DAZEL_NETWORK", DEFAULT_NETWORK),
                 run_deps=config.get("DAZEL_RUN_DEPS", DEFAULT_RUN_DEPS),
                 docker_compose_file=config.get("DAZEL_DOCKER_COMPOSE_FILE",
                                                DEFAULT_DOCKER_COMPOSE_FILE),
+                docker_compose_command=config.get("DAZEL_DOCKER_COMPOSE_COMMAND",
+                                                  DEFAULT_DOCKER_COMPOSE_COMMAND),
                 docker_compose_project_name=config.get("DAZEL_DOCKER_COMPOSE_PROJECT_NAME",
                                                        DEFAULT_DOCKER_COMPOSE_PROJECT_NAME),
                 docker_compose_services=config.get("DAZEL_DOCKER_COMPOSE_SERVICES",
@@ -99,23 +132,38 @@ class DockerInstance:
                                                   DEFAULT_BAZEL_USER_OUTPUT_ROOT),
                 docker_run_privileged=config.get("DAZEL_DOCKER_RUN_PRIVILEGED",
                                                  DEFAULT_DOCKER_RUN_PRIVILEGED),
-                dazel_run_file=config.get("DAZEL_RUN_FILE", DAZEL_RUN_FILE))
+                docker_machine=config.get("DAZEL_DOCKER_MACHINE",
+                                          DEFAULT_DOCKER_MACHINE),
+                dazel_run_file=config.get("DAZEL_RUN_FILE", DAZEL_RUN_FILE),
+                workspace_hex=config.get("DAZEL_WORKSPACE_HEX",
+                                          DEFAULT_WORKSPACE_HEX),
+                delegated_volume=config.get("DAZEL_DELEGATED_VOLUME", "DEFAULT_DELEGATED_VOLUME"),
+                user=config.get("DAZEL_USER", DEFAULT_USER),
+                docker_build_args=config.get("DAZEL_DOCKER_BUILD_ARGS", DEFAULT_DOCKER_BUILD_ARGS),
+        )
 
     def send_command(self, args):
-        command = "docker exec -i %s %s %s %s %s %s %s" % (
+        term_size = shutil.get_terminal_size()
+        command = "%s exec -i -e COLUMNS=%s -e LINES=%s -e TERM=%s %s %s %s %s %s %s %s %s" % (
+            self.docker_command,
+            term_size.columns, term_size.lines,
+            os.environ.get("TERM", ""),
             "-t" if sys.stdout.isatty() else "",
             "--privileged" if self.docker_run_privileged else "",
+            ("--user=%s" % self.user
+             if self.user else ""),
             self.instance_name,
             self.command,
             ("--bazelrc=%s" % self.bazel_rc_file
              if self.bazel_rc_file and self.command else ""),
             ("--output_user_root=%s --output_base=%s" % (
                 TEMP_BAZEL_OUTPUT_USER_ROOT, self.bazel_output_base)
-             if self.command and self.bazel_output_base
+             if self.command and self.bazel_output_base and not self.user
              else  "--output_user_root=%s" % self.bazel_user_output_root
                    if self.command and self.bazel_user_output_root
                    else ""),
             '"%s"' % '" "'.join(args))
+        command = self._with_docker_machine(command)
         return os.WEXITSTATUS(os.system(command))
 
     def start(self):
@@ -124,7 +172,7 @@ class DockerInstance:
 
         # Verify that the docker executable exists.
         if not self._docker_exists():
-            print ("ERROR: Docker executable could not be found!")
+            logger.error("ERROR: Docker executable could not be found!")
             return 1
 
         # Build or pull the relevant dazel image.
@@ -148,7 +196,7 @@ class DockerInstance:
 
             # Setup the network if necessary.
             if not self._network_exists():
-                print ("Creating network: '%s'" % self.network)
+                logger.info("Creating network: '%s'" % self.network)
                 rc = self._start_network()
             if rc:
                 return rc
@@ -163,7 +211,9 @@ class DockerInstance:
 
     def is_running(self):
         """Checks if the container is currently running."""
-        command = "docker ps | grep \"\\<%s\\>\" >/dev/null 2>&1" % (self.instance_name)
+        command = "%s ps | grep \"\\<%s\\>\" >/dev/null 2>&1" % (
+            self.docker_command, self.instance_name)
+        command = self._with_docker_machine(command)
 
         # If we have a directory, make sure the running container is mapped to
         # the same one (if not we need to create a new container mapped to the
@@ -180,15 +230,19 @@ class DockerInstance:
         if self.network:
             command += (" && docker inspect \"%s\" | grep '\"NetworkMode\": \"%s\"' >/dev/null 2>&1" %
                         (self.instance_name, self.network))
-            
-        rc = os.system(command)
+
+        rc = self._run_silent_command(command)
         return (rc == 0)
 
+    def _run_silent_command(self, command):
+        return subprocess.call(command, stdout=sys.stderr, shell=True)
+
     def _image_exists(self):
         """Checks if the dazel image exists in the local repository."""
-        command = "docker images | grep \"\\<%s/%s\\>\" >/dev/null 2>&1" % (
-            self.repository, self.image_name)
-        rc = os.system(command)
+        command = "%s images | grep \"\\<%s/%s\\>\" >/dev/null 2>&1" % (
+            self.docker_command, self.repository, self.image_name)
+        command = self._with_docker_machine(command)
+        rc = self._run_silent_command(command)
         return (rc == 0)
 
     def _build(self):
@@ -196,23 +250,26 @@ class DockerInstance:
         if not os.path.exists(self.dockerfile):
             raise RuntimeError("No Dockerfile to build the dazel image from.")
 
-        command = "docker build -t %s/%s -f %s %s" % (
-            self.repository, self.image_name, self.dockerfile, self.directory)
-        return os.system(command)
+        command = "%s build %s -t %s/%s -f %s %s" % (
+            self.docker_command, self.docker_build_args, self.repository, self.image_name, self.dockerfile, self.directory)
+        command = self._with_docker_machine(command)
+        return self._run_silent_command(command)
 
     def _pull(self):
         """Pulls the relevant image from the dockerhub repository."""
         if not self.repository:
             raise RuntimeError("No repository to pull the dazel image from.")
 
-        command = "docker pull %s/%s" % (self.repository, self.image_name)
-        return os.system(command)
+        command = "%s pull %s/%s" % (self.docker_command, self.repository, self.image_name)
+        command = self._with_docker_machine(command)
+        return self._run_silent_command(command)
 
     def _network_exists(self):
         """Checks if the network we need to use exists."""
-        command = "docker network ls | grep \"\\<%s\\>\" >/dev/null 2>&1" % (
-            self.network)
-        rc = os.system(command)
+        command = "%s network ls | grep \"\\<%s\\>\" >/dev/null 2>&1" % (
+            self.docker_command, self.network)
+        command = self._with_docker_machine(command)
+        rc = self._run_silent_command(command)
         return (rc == 0)
 
     def _start_network(self):
@@ -220,8 +277,9 @@ class DockerInstance:
         if not self.network:
             return 0
 
-        command = "docker network create %s" % self.network
-        return os.system(command)
+        command = "%s network create %s" % (self.docker_command, self.network)
+        command = self._with_docker_machine(command)
+        return self._run_silent_command(command)
 
     def _start_run_deps(self):
         """Starts the containers that are marked as runtime dependencies."""
@@ -230,6 +288,7 @@ class DockerInstance:
                 instance_name=run_dep_name,
                 image_name=run_dep_image,
                 run_command=None,
+                docker_command=None,
                 dockerfile=None,
                 repository=None,
                 directory=None,
@@ -239,15 +298,17 @@ class DockerInstance:
                 network=self.network,
                 run_deps=None,
                 docker_compose_file=None,
+                docker_compose_command=None,
                 docker_compose_project_name=None,
                 docker_compose_services=None,
                 bazel_rc_file=None,
                 bazel_user_output_root=None,
                 docker_run_privileged=self.docker_run_privileged,
+                docker_machine=self.docker_machine,
                 dazel_run_file=None)
             if not run_dep_instance.is_running():
-                print ("Starting run dependency: '%s' (name: '%s')" %
-                       (run_dep_image, run_dep_name))
+                logger.info ("Starting run dependency: '%s' (name: '%s')" %
+                             (run_dep_image, run_dep_name))
                 run_dep_instance._run_container()
 
     def _start_compose_services(self):
@@ -255,40 +316,46 @@ class DockerInstance:
         if not self.docker_compose_file:
             return 0
 
-        command = "COMPOSE_PROJECT_NAME=%s docker-compose -f %s pull --ignore-pull-failures %s" % (
-            self.docker_compose_project_name, self.docker_compose_file,
+        command = "COMPOSE_PROJECT_NAME=%s %s -f %s pull --ignore-pull-failures %s" % (
+            self.docker_compose_project_name, self.docker_compose_command, self.docker_compose_file,
             self.docker_compose_services)
-        command += " && COMPOSE_PROJECT_NAME=%s docker-compose -f %s build %s" % (
-            self.docker_compose_project_name, self.docker_compose_file,
+        command += " && COMPOSE_PROJECT_NAME=%s %s -f %s build %s" % (
+            self.docker_compose_project_name, self.docker_compose_command, self.docker_compose_file,
             self.docker_compose_services)
-        command += " && COMPOSE_PROJECT_NAME=%s docker-compose -f %s up --force-recreate -d %s" % (
-            self.docker_compose_project_name, self.docker_compose_file,
+        command += " && COMPOSE_PROJECT_NAME=%s %s -f %s up --force-recreate -d %s" % (
+            self.docker_compose_project_name, self.docker_compose_command, self.docker_compose_file,
             self.docker_compose_services)
-        return os.system(command)
+        command = self._with_docker_machine(command)
+        return self._run_silent_command(command)
 
     def _run_container(self):
         """Runs the container itself."""
-        print ("Starting docker container '%s'..." % self.instance_name)
-        command = "docker stop %s >/dev/null 2>&1 ; " % (self.instance_name)
-        command += "docker rm %s >/dev/null 2>&1 ; " % (self.instance_name)
-        command += "docker run -id --name=%s %s %s %s %s %s %s%s %s" % (
+        logger.info("Starting docker container '%s'..." % self.instance_name)
+        command = "%s stop %s >/dev/null 2>&1 ; " % (self.docker_command, self.instance_name)
+        command += "%s rm %s >/dev/null 2>&1 ; " % (self.docker_command, self.instance_name)
+        command += "%s run -id --name=%s %s %s %s %s %s %s %s %s%s %s" % (
+            self.docker_command,
             self.instance_name,
             "--privileged" if self.docker_run_privileged else "",
+            ("--user=%s" % self.user
+             if self.user else ""),
             ("-w %s" % os.path.realpath(self.directory)) if self.directory else "",
             self.volumes,
             self.ports,
+            self.env_vars,
             ("--net=%s" % self.network) if self.network else "",
             ("%s/" % self.repository) if self.repository else "",
             self.image_name,
             self.run_command if self.run_command else "")
-        rc = os.system(command)
+        command = self._with_docker_machine(command)
+        rc = self._run_silent_command(command)
         if rc:
             return rc
 
         # Touch the dazel run file to change the timestamp.
         if self.dazel_run_file:
             open(self.dazel_run_file, "w").write(self.instance_name + "\n")
-            print ("Done.")
+            logger.info("Done.")
 
         return rc
 
@@ -302,7 +369,7 @@ class DockerInstance:
         # DAZEL_VOLUMES can be a python iterable or a comma-separated string.
         if isinstance(volumes, str):
             volumes = [v.strip() for v in volumes.split(",")]
-        elif volumes and not isinstance(volumes, types.Iterable):
+        elif volumes and not isinstance(volumes, collections.Iterable):
             raise RuntimeError("DAZEL_VOLUMES must be comma-separated string "
                                "or python iterable of strings")
 
@@ -326,10 +393,9 @@ class DockerInstance:
         # Add the bazel user output directory if it exists, or the real bazelout
         # directory if it does.
         if self.bazel_user_output_root:
-            workspace_hex_digest = hashlib.md5(real_directory.encode("ascii")).hexdigest()
             self.bazel_output_base = os.path.realpath(
                 os.path.join(self.bazel_user_output_root,
-                             workspace_hex_digest))
+                             self.workspace_hex_digest))
 
             user_output_paths = (DEFAULT_BAZEL_USER_OUTPUT_PATHS +
                                  [os.path.basename(real_directory)])
@@ -339,10 +405,11 @@ class DockerInstance:
                                user_output_path))
               if not os.path.isdir(real_user_output_path):
                   os.makedirs(real_user_output_path)
-              volumes += ["%s:%s" % (real_user_output_path,
-                                     real_user_output_path)]
+              volumes += ["%s:%s%s" % (real_user_output_path,
+                                       real_user_output_path,
+                                       self.delegated_volume_flag)]
         elif real_bazelout:
-            volumes += ["%s:%s" % (real_bazelout, real_bazelout)]
+            volumes += ["%s:%s%s" % (real_bazelout, real_bazelout, self.delegated_volume_flag)]
             self.bazel_output_base = real_bazelout
 
         # Make sure the path exists on the host.
@@ -354,7 +421,7 @@ class DockerInstance:
 
     def _add_ports(self, ports):
         """Add the given ports to the run string."""
-        # This can only be intentional in code, so ignore None volumes.
+        # This can only be intentional in code, so disregard.
         self.ports = ""
         if not ports:
             return
@@ -362,13 +429,30 @@ class DockerInstance:
         # DAZEL_PORTS can be a python iterable or a comma-separated string.
         if isinstance(ports, str):
             ports = [p.strip() for p in ports.split(",")]
-        elif ports and not isinstance(ports, types.Iterable):
+        elif ports and not isinstance(ports, collections.Iterable):
             raise RuntimeError("DAZEL_PORTS must be comma-separated string "
                                "or python iterable of strings")
 
-        # Find the real source and output directories.
+        # calculate the ports string
         self.ports = '-p "%s"' % '" -p "'.join(ports)
 
+    def _add_env_vars(self, env_vars):
+        """Add the given env vars to the run string."""
+        # This can only be intentional in code, so disregard.
+        self.env_vars = ""
+        if not env_vars:
+            return
+
+        # DAZEL_ENV_VARS can be a python iterable or a comma-separated string.
+        if isinstance(env_vars, str):
+            env_vars = [p.strip() for p in env_vars.split(",")]
+        elif env_vars and not isinstance(env_vars, collections.Iterable):
+            raise RuntimeError("DAZEL_ENV_VARS must be comma-separated string "
+                               "or python iterable of strings")
+
+        # calculate the env string
+        self.env_vars = '-e "%s"' % '" -e "'.join(env_vars)
+
     def _add_run_deps(self, run_deps):
         """Adds the necessary runtime container dependencies to launch."""
         # This can only be intentional in code, so disregard.
@@ -379,7 +463,7 @@ class DockerInstance:
         # DAZEL_RUN_DEPS can be a python iterable or a comma-separated string.
         if isinstance(run_deps, str):
             run_deps = [rd.strip() for rd in run_deps.split(",")]
-        elif run_deps and not isinstance(run_deps, types.Iterable):
+        elif run_deps and not isinstance(run_deps, collections.Iterable):
             raise RuntimeError("DAZEL_RUN_DEPS must be comma-separated string "
                                "or python iterable of strings")
 
@@ -400,7 +484,7 @@ class DockerInstance:
         # comma-separated string.
         if isinstance(docker_compose_services, str):
             docker_compose_services = [s.strip() for s in docker_compose_services.split(",")]
-        elif docker_compose_services and not isinstance(docker_compose_services, types.Iterable):
+        elif docker_compose_services and not isinstance(docker_compose_services, collections.Iterable):
             raise RuntimeError("DAZEL_DOCKER_COMPOSE_SERVICES must be comma-separated string "
                                "or python iterable of strings")
 
@@ -409,18 +493,23 @@ class DockerInstance:
 
     def _docker_exists(self):
         """Checks if the basic docker executable exists."""
-        return self._command_exists("docker")
+        return self._command_exists(self.docker_command)
 
     def _docker_compose_exists(self):
         """Checks if the docker-compose executable exists."""
-        return self._command_exists("docker-compose")
+        return self._command_exists(self.docker_compose_command)
 
     def _command_exists(self, cmd):
         """Checks if a command exists on the system."""
         command = "which %s >/dev/null 2>&1" % (cmd)
-        rc = os.system(command)
+        rc = self._run_silent_command(command)
         return (rc == 0)
 
+    def _with_docker_machine(self, cmd):
+        if self.docker_machine is None or not self._command_exists("docker-machine"):
+            return cmd
+        return "eval $(docker-machine env %s) && (%s)" % (self.docker_machine, cmd)
+
     @classmethod
     def _config_from_file(cls):
         """Creates a configuration from a .dazelrc file."""
@@ -478,4 +567,3 @@ def main():
 
 if __name__ == "__main__":
     sys.exit(main())
-
